/*
==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

Dia-Args

Copyright (C) 2018-2019, 2021-2024  Anonymous

There are several releases over multiple years,
they are listed as ranges, such as: "2018-2019".

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

::--::--::--::--::--::--::--::--::--::--::--::--::--::--::--::--
*/

//! # Some kit

use {
    core::{
        fmt::Debug,
        str::FromStr,
    },
    std::io::{self, BufRead, Error, ErrorKind, Write},
    crate::Result,
};

/// # Reads a line from stdin, trims and converts it to `T`
pub fn read_line<T>() -> Result<T> where T: FromStr, <T as FromStr>::Err: Debug {
    let stdin = io::stdin();
    let mut stdin = stdin.lock();

    let mut buf = String::with_capacity(64);
    stdin.read_line(&mut buf)?;

    let buf = buf.trim();
    T::from_str(buf).map_err(|err| Error::new(ErrorKind::InvalidData, format!("{:?} is invalid: {:?}", buf, err)))
}

/// # Locks stdout and writes to it
///
/// - For simplicity, any error will be caught and printed out via `eprintln!()`.
/// - The function will flush stdout when done.
pub fn lock_write_out<B>(bytes: B) where B: AsRef<[u8]> {
    let stdout = io::stdout();
    let mut stdout = stdout.lock();
    if let Err(err) = (|| {
        stdout.write_all(bytes.as_ref())?;
        stdout.flush()
    })() {
        eprintln!("{}", __!("{}", err));
    }
}

/// # Locks stderr and writes to it
///
/// - For simplicity, any error will be caught and printed out via `eprintln!()`.
/// - The function will flush stderr when done.
pub fn lock_write_err<B>(bytes: B) where B: AsRef<[u8]> {
    let stderr = io::stderr();
    let mut stderr = stderr.lock();
    if let Err(err) = (|| {
        stderr.write_all(bytes.as_ref())?;
        stderr.flush()
    })() {
        eprintln!("{}", __!("{}", err));
    }
}
